from unittest import TestCase
from cilantro_ee.nodes.masternode.transaction_batcher import TransactionBatcher, NewTransactionBatcher, MN_TX_PUB_PORT
from cilantro_ee.core.crypto.wallet import Wallet
import zmq
import zmq.asyncio
import asyncio

from tests import random_txs

from cilantro_ee.constants.zmq_filters import TRANSACTION_FILTER
from cilantro_ee.core.messages.message import Message
from cilantro_ee.core.messages.message_type import MessageType


class TestTransactionBatcher(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.loop = asyncio.get_event_loop()

    def tearDown(self):
        self.ctx.destroy()
        self.loop.close()

    def test_init(self):
        w = Wallet()
        TransactionBatcher('127.0.0.1', w.signing_key().hex())

    def test_wait_until_ready_ends_when_ready_is_true(self):
        w = Wallet()
        t = TransactionBatcher('0.0.0.0', w.signing_key().hex())

        async def set_ready():
            t._ready = True

        tasks = asyncio.gather(
            t._wait_until_ready(),
            set_ready()
        )

        t.loop.run_until_complete(tasks)


async def stop_server(s, timeout):
    await asyncio.sleep(timeout)
    s.stop()


class TestNewTransactionBatcher(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.loop = asyncio.get_event_loop()

    def tearDown(self):
        self.ctx.destroy()
        #self.loop.close()

    def test_init(self):
        w = Wallet()
        t = NewTransactionBatcher(publisher_ip='127.0.0.1', wallet=w, ctx=self.ctx)
        print(t.ipc_id)

    def test_start_and_stop_works(self):
        w = Wallet()
        t = NewTransactionBatcher(publisher_ip='127.0.0.1', wallet=w, ctx=self.ctx)

        tasks = asyncio.gather(
            t.start(),
            stop_server(t, 0.2)
        )

        self.loop.run_until_complete(tasks)

    def test_input_hash_inbox_ready_sets_ready_on_tx_batcher(self):
        w1 = Wallet()
        t = NewTransactionBatcher(publisher_ip='127.0.0.1', wallet=w1, ctx=self.ctx)

        self.assertFalse(t.ready)

        w2 = Wallet()

        message = Message.get_signed_message_packed_2(wallet=w2,
                                                      msg_type=MessageType.READY)

        async def get(msg):
            await asyncio.sleep(0.1)
            socket = self.ctx.socket(zmq.DEALER)
            socket.connect('ipc:///tmp/masternode-input-hash-inbox')

            await socket.send(msg)

        tasks = asyncio.gather(
            t.input_hash_inbox.serve(),
            get(message),
            stop_server(t.input_hash_inbox, 0.2),
        )

        self.loop.run_until_complete(tasks)

        self.assertTrue(t.ready)

    def test_input_hash_inbox_burn_input_hashes_removes_batch_ids_from_rate_limiter(self):
        w1 = Wallet()
        t = NewTransactionBatcher(publisher_ip='127.0.0.1', wallet=w1, ctx=self.ctx)

        t.batcher.add_batch_id(b'hello_there')

        self.assertListEqual(t.batcher.sent_batch_ids, [b'hello_there'])

        w2 = Wallet()
        message = Message.get_signed_message_packed_2(wallet=w2,
                                                      msg_type=MessageType.BURN_INPUT_HASHES,
                                                      inputHashes=[b'hello_there'])

        async def get(msg):
            await asyncio.sleep(0.1)
            socket = self.ctx.socket(zmq.DEALER)
            socket.connect('ipc:///tmp/masternode-input-hash-inbox')

            await socket.send(msg)

        tasks = asyncio.gather(
            t.input_hash_inbox.serve(),
            get(message),
            stop_server(t.input_hash_inbox, 0.2),
        )

        self.loop.run_until_complete(tasks)

        self.assertListEqual(t.batcher.sent_batch_ids, [])

    def test_compose_transactions_publishes_to_subscriber(self):
        class MockQueue:
            def __init__(self):
                self.q = []

            def get(self):
                return self.q.pop(0)

            def put_nowait(self, obj):
                self.q.append(obj)

            def qsize(self):
                return len(self.q)

        w1 = Wallet()
        tx = random_txs.random_packed_tx(0, processor=w1.verifying_key(), give_stamps=True)
        t = NewTransactionBatcher(publisher_ip='127.0.0.1', wallet=w1, ctx=self.ctx, queue=MockQueue())

        t.queue.put_nowait((0, tx))
        t.batcher.max_txn_submission_delay = 0

        sub_socket = self.ctx.socket(zmq.SUB)
        sub_socket.connect(f'tcp://{t.publisher_ip}:{MN_TX_PUB_PORT}')
        sub_socket.setsockopt_string(zmq.SUBSCRIBE, '')

        t.running = True

        async def get(socket):
            msg = await socket.recv()
            return msg

        tasks = asyncio.gather(
            t.compose_transactions(),
            get(sub_socket),
            stop_server(t, 0.3),
        )

        results = self.loop.run_until_complete(tasks)

        msg_type, unpacked, _, _, _ = Message.unpack_message_2(results[1])

        self.assertDictEqual(unpacked.transactions[0].to_dict(), tx.to_dict())



from cilantro_ee.services.storage.vkbook import PhoneBook
from cilantro_ee.constants import conf
from cilantro_ee.services.overlay.network import Network
from cilantro_ee.constants.ports import DHT_PORT, EVENT_PORT
from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.core.sockets.services import AsyncInbox, SocketStruct, get
from cilantro_ee.core.messages.message import Message
from cilantro_ee.core.messages.message_type import MessageType
import zmq.asyncio
from cilantro_ee.core.crypto.wallet import Wallet
import time


class OverlayServer(AsyncInbox):
    def __init__(self, socket_id: SocketStruct,
                 wallet: Wallet,
                 ctx: zmq.Context,
                 ip=conf.HOST_IP,
                 peer_service_port=DHT_PORT,
                 event_publisher_port=EVENT_PORT,
                 bootnodes=conf.BOOTNODES,
                 initial_mn_quorum=PhoneBook.num_boot_masternodes,
                 initial_del_quorum=PhoneBook.num_boot_delegates,
                 mn_to_find=PhoneBook.masternodes,
                 del_to_find=PhoneBook.delegates,
                 linger=2000,
                 poll_timeout=2000):

        super().__init__(socket_id, wallet, ctx, linger, poll_timeout)

        self.network_address = 'tcp://{}:{}'.format(conf.HOST_IP, DHT_PORT)

        self.network = Network(wallet=self.wallet,
                               ctx=self.ctx,
                               ip=ip,
                               peer_service_port=peer_service_port,
                               event_publisher_port=event_publisher_port,
                               bootnodes=bootnodes,
                               initial_mn_quorum=initial_mn_quorum,
                               initial_del_quorum=initial_del_quorum,
                               mn_to_find=mn_to_find,
                               del_to_find=del_to_find)

        self.log = get_logger('Overlay.Server')

    async def serve(self):
        await self.network.start()
        await super().serve()

    async def return_bad_request(self, _id):
        reply = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                    msg_type=MessageType.BAD_REQUEST,
                                                    timestamp=int(time.time()))
        await self.return_msg(_id, reply)

    async def handle_msg(self, _id, msg):
        print(msg)
        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(message=msg)
        if msg_type == MessageType.IP_FOR_VK_REQUEST:
            response = await self.network.find_node(vk_to_find=msg.vk.hex())

            ip = response.get(msg.vk.hex())
            if ip is not None:
                reply = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                            msg_type=MessageType.IP_FOR_VK_REPLY,
                                                            ip=ip.encode())
                await self.return_msg(_id, reply)
            else:
                await self.return_bad_request(_id)
        else:
            await self.return_bad_request(_id)


class OverlayClient:
    def __init__(self, wallet: Wallet, ctx: zmq.Context, overlay_server_socket: SocketStruct):
        self.wallet = wallet
        self.ctx = ctx
        self.overlay_server_socket = overlay_server_socket

    async def get_ip_for_vk(self, vk: bytes):
        req = Message.get_signed_message_packed_2(wallet=self.wallet,
                                                  msg_type=MessageType.IP_FOR_VK_REQUEST,
                                                  vk=vk)

        resp = await get(self.overlay_server_socket, msg=req, ctx=self.ctx, dealer=True)

        if resp is not None:
            msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(message=resp)

            return msg.ip.decode()



from cilantro_ee.messages.block_data.state_update import BlockIndexReply
from unittest import TestCase


class TestBlockIndexReply(TestCase):

    def _build_reply(self, return_data=False):
        mn1 = 'ABCD' * 16
        mn2 = 'DCBA' * 16
        mn3 = 'AABB' * 16

        b1 = 'A' * 64
        b2 = 'B' * 64
        b3 = 'C' * 64

        data = [[b1, 1, [mn1,]], [b2, 2, [mn2, mn1]], [b3, 3, [mn1, mn2, mn3]]]
        if return_data:
            return BlockIndexReply.create(data), data
        else:
            return BlockIndexReply.create(data)

    def test_create(self):
        reply, data = self._build_reply(return_data=True)
        self.assertEqual(reply.indices, data)

    def test_create_with_empty(self):
        reply = BlockIndexReply.create([])
        self.assertEqual(reply.indices, [])

    def test_serialization(self):
        reply = self._build_reply()
        self.assertEqual(reply, BlockIndexReply.from_bytes(reply.serialize()))
